import abc
import dataclasses
import os
from typing import Any, ClassVar, Dict, Generator, Generic, List, Optional, Self, TypeVar

from semver import Version
from typing_extensions import override

from aind_behavior_core_analysis import _typing


class DataStream(abc.ABC, Generic[_typing.TData, _typing.TReaderParams]):
    """Abstract base class for all data streams.

    Provides a generic interface for data reading operations with configurable parameters
    and hierarchical organization.

    Args:
        name: Name identifier for the data stream.
        description: Optional description of the data stream.
        reader_params: Optional parameters for the data reader.
        **kwargs: Additional keyword arguments.

    Attributes:
        _is_collection: Class variable indicating if this is a collection of data streams.

    Raises:
        ValueError: If name contains '::' characters which are reserved for path resolution.
    """

    _is_collection: ClassVar[bool] = False

    def __init__(
        self: Self,
        name: str,
        *,
        description: Optional[str] = None,
        reader_params: Optional[_typing.TReaderParams] = None,
        **kwargs,
    ) -> None:
        if "::" in name:
            raise ValueError("Name cannot contain '::' character.")
        self._name = name

        self._description = description
        self._reader_params = reader_params if reader_params is not None else _typing.UnsetParams
        self._data = _typing.UnsetData
        self._parent: Optional["DataStream"] = None

    @property
    def name(self) -> str:
        """Get the name of the data stream.

        Returns:
            str: Name identifier of the data stream.
        """
        return self._name

    @property
    def resolved_name(self) -> str:
        """Get the full hierarchical name of the data stream.

        Generates a path-like name showing the stream's position in the hierarchy,
        using '::' as a separator between parent and child names.

        Returns:
            str: The fully resolved name including all parent names.
        """
        builder = self.name
        d = self
        while d._parent is not None:
            builder = f"{d._parent.name}::{builder}"
            d = d._parent
        return builder

    @property
    def description(self) -> Optional[str]:
        """Get the description of the data stream.

        Returns:
            Optional[str]: Description of the data stream, or None if not provided.
        """
        return self._description

    @property
    def parent(self) -> Optional["DataStream"]:
        """Get the parent data stream.

        Returns:
            Optional[DataStream]: Parent data stream, or None if this is a root stream.
        """
        return self._parent

    @property
    def is_collection(self) -> bool:
        """Check if this data stream is a collection of other streams.

        Returns:
            bool: True if this is a collection stream, False otherwise.
        """
        return self._is_collection

    _reader: _typing.IReader[_typing.TData, _typing.TReaderParams] = _typing.UnsetReader

    make_params = NotImplementedError("make_params is not implemented for DataStream.")

    @property
    def reader_params(self) -> _typing.TReaderParams:
        """Get the parameters for the data reader.

        Returns:
            TReaderParams: Parameters for the data reader.
        """
        return self._reader_params

    def read(self, reader_params: Optional[_typing.TReaderParams] = None) -> _typing.TData:
        """Read data using the configured reader.

        Args:
            reader_params: Optional parameters to override the default reader parameters.

        Returns:
            TData: Data read from the source.

        Raises:
            ValueError: If reader parameters are not set.
        """
        reader_params = reader_params if reader_params is not None else self._reader_params
        if _typing.is_unset(reader_params):
            raise ValueError("Reader parameters are not set. Cannot read data.")
        return self._reader(reader_params)

    def bind_reader_params(self, params: _typing.TReaderParams) -> Self:
        """Bind reader parameters to the data stream.

        Args:
            params: Parameters to bind to the data stream's reader.

        Returns:
            Self: The data stream instance for method chaining.

        Raises:
            ValueError: If reader parameters have already been set.
        """
        if not _typing.is_unset(self._reader_params):
            raise ValueError("Reader parameters are already set. Cannot bind again.")
        self._reader_params = params
        return self

    def at(self, name: str) -> "DataStream":
        """Get a child data stream by name.

        Args:
            name: Name of the child data stream to retrieve.

        Returns:
            DataStream: The child data stream with the given name.

        Raises:
            NotImplementedError: If the data stream does not support child access.
        """
        raise NotImplementedError("This method is not implemented for DataStream.")

    def __getitem__(self, name: str) -> "DataStream":
        """Get a child data stream by name using dictionary-like syntax.

        Args:
            name: Name of the child data stream to retrieve.

        Returns:
            DataStream: The child data stream with the given name.
        """
        return self.at(name)

    @property
    def has_data(self) -> bool:
        """Check if the data stream has loaded data.

        Returns:
            bool: True if data has been loaded, False otherwise.
        """
        return not _typing.is_unset(self._data)

    @property
    def data(self) -> _typing.TData:
        """Get the loaded data.

        Returns:
            TData: The loaded data.

        Raises:
            ValueError: If data has not been loaded yet.
        """
        if not self.has_data:
            raise ValueError("Data has not been loaded yet.")
        return self._data

    def load(self) -> Self:
        """Load data into the data stream.

        Reads data from the source and stores it in the data stream.

        Returns:
            Self: The data stream instance for method chaining.
        """
        self._data = self.read()
        return self

    def __str__(self):
        """Generate a string representation of the data stream.

        Returns:
            str: String representation showing the stream's type, name, description,
                 reader parameters, and data type if loaded.
        """
        return (
            f"DataStream("
            f"stream_type={self.__class__.__name__}, "
            f"name={self._name}, "
            f"description={self._description}, "
            f"reader_params={self._reader_params}, "
            f"data_type={self._data.__class__.__name__ if self.has_data else 'Not Loaded'}"
        )

    def __iter__(self) -> Generator["DataStream", None, None]:
        """Iterator interface for data streams.

        For non-collection data streams, this yields nothing.

        Yields:
            DataStream: Child data streams (none for base DataStream).
        """
        yield

    def load_all(self, strict: bool = False) -> list[tuple["DataStream", Exception], None, None]:
        """Recursively load this data stream and all child streams.

        Performs depth-first traversal to load all streams in the hierarchy.

        Args:
            strict: If True, raises exceptions immediately; otherwise collects and returns them.

        Returns:
            list: List of tuples containing streams and exceptions that occurred during loading.

        Raises:
            Exception: If strict is True and an exception occurs during loading.
        """
        self.load()
        exceptions = []
        for stream in self:
            if stream is None:
                continue
            try:
                stream.load_all(strict=strict)
            except Exception as e:
                if strict:
                    raise e
                exceptions.append((stream, e))
        return exceptions


TDataStream = TypeVar("TDataStream", bound=DataStream[Any, Any])


class DataStreamCollectionBase(
    DataStream[List[TDataStream], _typing.TReaderParams],
    Generic[TDataStream, _typing.TReaderParams],
):
    """Base class for collections of data streams.

    Provides functionality for managing and accessing multiple child data streams.

    Args:
        name: Name identifier for the collection.
        description: Optional description of the collection.
        reader_params: Optional parameters for the reader.
        **kwargs: Additional keyword arguments.
    """

    _is_collection: ClassVar[bool] = True

    def __init__(
        self: Self,
        name: str,
        *,
        description: Optional[str] = None,
        reader_params: Optional[_typing.TReaderParams] = None,
        **kwargs,
    ) -> None:
        super().__init__(name=name, description=description, reader_params=reader_params, **kwargs)
        self._hashmap: Dict[str, TDataStream] = {}
        self._update_hashmap()

    def _update_hashmap(self) -> None:
        """Update the internal hashmap of child data streams.

        Validates that all child streams have unique names and updates the lookup table.

        Raises:
            ValueError: If duplicate names are found among child streams.
        """
        if not self.has_data:
            return
        stream_keys = [stream.name for stream in self.data]
        duplicates = [name for name in stream_keys if stream_keys.count(name) > 1]
        if duplicates:
            raise ValueError(f"Duplicate names found in the data stream collection: {set(duplicates)}")
        self._hashmap = {stream.name: stream for stream in self.data}
        self._update_parent_references()
        return

    def _update_parent_references(self) -> None:
        """Update parent references for all child data streams.

        Sets this collection as the parent for all child streams.
        """
        for stream in self._hashmap.values():
            stream._parent = self

    @override
    def load(self):
        """Load data for this collection.

        Overrides the base method to add validation that loaded data is a list of DataStreams.

        Returns:
            Self: The collection instance for method chaining.

        Raises:
            ValueError: If loaded data is not a list of DataStreams.
        """
        super().load()
        if not isinstance(self._data, list):
            self._data = _typing.UnsetData
            raise ValueError("Data must be a list of DataStreams.")
        self._update_hashmap()
        return self

    @override
    def at(self, name: str) -> TDataStream:
        """Get a child data stream by name.

        Args:
            name: Name of the child data stream to retrieve.

        Returns:
            TDataStream: The child data stream with the given name.

        Raises:
            ValueError: If data has not been loaded yet.
            KeyError: If no child stream with the given name exists.
        """
        if not self.has_data:
            raise ValueError("data streams have not been read yet. Cannot access data streams.")
        if name in self._hashmap:
            return self._hashmap[name]
        else:
            raise KeyError(f"Stream with name: '{name}' not found in data streams.")

    def __str__(self: Self) -> str:
        """Generate a formatted table representation of the collection.

        Returns:
            str: Formatted table showing child streams, their types, and load status.
        """
        table = []
        table.append(["Stream Name", "Stream Type", "Is Loaded"])
        table.append(["-" * 20, "-" * 20, "-" * 20])

        if not self.has_data:
            return f"{self.__class__.__name__} has not been loaded yet."

        for key, value in self._hashmap.items():
            table.append(
                [key, value.data.__class__.__name__ if value.has_data else "Unknown", "Yes" if value.has_data else "No"]
            )

        max_lengths = [max(len(str(row[i])) for row in table) for i in range(len(table[0]))]

        formatted_table = []
        for row in table:
            formatted_row = [str(cell).ljust(max_lengths[i]) for i, cell in enumerate(row)]
            formatted_table.append(formatted_row)

        table_str = ""
        for row in formatted_table:
            table_str += " | ".join(row) + "\n"

        return table_str

    def __iter__(self) -> Generator[DataStream, None, None]:
        """Iterator for all child data streams, including nested collections.

        Implements a depth-first traversal of the stream hierarchy.

        Yields:
            DataStream: Child data streams.
        """
        for value in self._hashmap.values():
            if isinstance(value, DataStream):
                yield value
            if isinstance(value, DataStreamCollectionBase):
                yield from value.__iter__()


class DataStreamCollection(DataStreamCollectionBase[DataStream, _typing.UnsetParamsType]):
    """Collection of data streams with direct initialization.

    A specialized collection where child streams are passed directly instead of being
    created by a reader function.

    Args:
        name: Name identifier for the collection.
        data_streams: List of child data streams to include.
        description: Optional description of the collection.
    """

    @override
    def __init__(
        self,
        name: str,
        data_streams: List[DataStream],
        *,
        description: Optional[str] = None,
    ) -> None:
        """Initializes a special DataStreamGroup where the data streams are passed directly, without a reader."""
        super().__init__(
            name=name,
            description=description,
            reader_params=_typing.UnsetParams,
        )
        self.bind_data_streams(data_streams)

    @staticmethod
    def parameters(*args, **kwargs) -> _typing.UnsetParamsType:
        """Parameters function to return UnsetParams.

        Returns:
            UnsetParamsType: Special unset parameters value.
        """
        return _typing.UnsetParams

    def _reader(self, *args, **kwargs) -> List[DataStream]:
        """Reader function that returns the pre-set data streams.

        Returns:
            List[DataStream]: The pre-set data streams.
        """
        return self._data

    @override
    def read(self, *args, **kwargs) -> List[DataStream]:
        """Read data from the collection.

        Returns:
            List[DataStream]: The pre-set data streams.

        Raises:
            ValueError: If data streams have not been set yet.
        """
        if not self.has_data:
            raise ValueError("Data streams have not been read yet.")
        return self._data

    def bind_data_streams(self, data_streams: List[DataStream]) -> Self:
        """Bind a list of data streams to the collection.

        Args:
            data_streams: List of data streams to include in the collection.

        Returns:
            Self: The collection instance for method chaining.

        Raises:
            ValueError: If data streams have already been set.
        """
        if self.has_data:
            raise ValueError("Data streams are already set. Cannot bind again.")
        self._data = data_streams
        self._update_hashmap()
        return self

    def add_stream(self, stream: DataStream) -> Self:
        """Add a new data stream to the collection.

        Args:
            stream: Data stream to add to the collection.

        Returns:
            Self: The collection instance for method chaining.

        Raises:
            KeyError: If a stream with the same name already exists.
        """
        if not self.has_data:
            self._data = [stream]
            self._update_hashmap()
            return self

        if stream.name in self._hashmap:
            raise KeyError(f"Stream with name: '{stream.name}' already exists in data streams.")

        self._data.append(stream)
        self._update_hashmap()
        return self

    def remove_stream(self, name: str) -> None:
        """Remove a data stream from the collection.

        Args:
            name: Name of the data stream to remove.

        Raises:
            ValueError: If data streams have not been set yet.
            KeyError: If no stream with the given name exists.
        """
        if not self.has_data:
            raise ValueError("Data streams have not been read yet. Cannot access data streams.")

        if name not in self._hashmap:
            raise KeyError(f"Data stream with name '{name}' not found in data streams.")
        self._data.remove(self._hashmap[name])
        self._update_hashmap()
        return

    @classmethod
    def from_data_stream(cls, data_stream: DataStream) -> Self:
        """Create a DataStreamCollection from a DataStream object.

        Factory method to convert a single data stream or collection into a DataStreamCollection.

        Args:
            data_stream: Source data stream to convert.

        Returns:
            DataStreamCollection: New collection containing the source stream's data.

        Raises:
            TypeError: If the source is not a DataStream.
            ValueError: If the source has not been loaded yet.
        """
        if not isinstance(data_stream, DataStream):
            raise TypeError("data_stream must be an instance of DataStream.")
        if not data_stream.has_data:
            raise ValueError("DataStream has not been loaded yet. Cannot create DataStreamCollection.")
        data = data_stream.data if data_stream.is_collection else [data_stream.data]
        return cls(name=data_stream.name, data_streams=data, description=data_stream.description)


class Dataset(DataStreamCollection):
    """A version-tracked collection of data streams.

    Extends DataStreamCollection by adding semantic versioning support.

    Args:
        name: Name identifier for the dataset.
        data_streams: List of data streams to include in the dataset.
        version: Semantic version string or Version object. Defaults to "0.0.0".
        description: Optional description of the dataset.
    """

    @override
    def __init__(
        self,
        name: str,
        data_streams: List[DataStream],
        *,
        version: str | Version = "0.0.0",
        description: Optional[str] = None,
    ) -> None:
        """Initializes a Dataset with a version and a list of data streams."""
        super().__init__(
            name=name,
            data_streams=data_streams,
            description=description,
        )
        self._version = self._parse_semver(version)

    @staticmethod
    def _parse_semver(version: str | Version) -> Version:
        """Parse a version string into a Version object.

        Args:
            version: Version string or object to parse.

        Returns:
            Version: Semantic version object.
        """
        if isinstance(version, str):
            return Version.parse(version)
        return version

    @property
    def version(self) -> Version:
        """Get the semantic version of the dataset.

        Returns:
            Version: Semantic version object.
        """
        return self._version


@dataclasses.dataclass
class FilePathBaseParam(abc.ABC):
    """Abstract base class for file-based reader parameters.

    Base parameter class for readers that access files by path.

    Attributes:
        path: Path to the file or directory to read from.
    """

    path: os.PathLike
